#!/usr/bin/env python3

import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from mpl_toolkits.mplot3d import Axes3D
from itertools import combinations
import os
from crazyKhoreia._3DIoU import _3DIoU
from crazyKhoreia.crazyKhoreia import crazyKhoreia

class multiDronFormation():
  def __init__(self, MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, MIN_DEPTH, MAX_DEPTH, boxShape, in_path, out_path, nmbr_drones):
    self.MAX_WIDTH, self.MAX_HEIGHT, self.MIN_WIDTH, self.MIN_HEIGHT, self.MIN_DEPTH, self.MAX_DEPTH, self.boxShape, self.in_path, self.out_path = MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, MIN_DEPTH, MAX_DEPTH, boxShape, in_path, out_path
    self.nmbr_drones = nmbr_drones

    self.ck = crazyKhoreia(self.MAX_WIDTH, self.MAX_HEIGHT,
                            self.MIN_WIDTH, self.MIN_HEIGHT, self.in_path)
    
    self.wayPoints = self.ck.get_waypoints()
    self.ck.plot_contour_inspection(self.wayPoints)

    cc = self.get_clusters()
    positions = self.get_idealPositions(cc)
    self.getIoUsppd(positions)

    self.save(positions, in_path, out_path)

    plt.show()

  def get_clusters(self):
    fig, ax0 = plt.subplots()
    kmeans = KMeans(n_clusters = self.nmbr_drones, random_state=0)
    y_pred = kmeans.fit_predict(np.array([self.wayPoints[:,1], self.wayPoints[:,2]]).T)
    cc = kmeans.cluster_centers_

    ax0.scatter(self.wayPoints[:,1], self.wayPoints[:,2], c=y_pred)
    ax0.plot(cc[:,0], cc[:,1], 'o', c='violet', label='Cluster centroids.')
    ax0.legend()
    ax0.set_title("KMeans clusters.")

    return cc

  def get_idealPositions(self, cc):
    positions = np.empty(shape=(0, 3))
    
    positions = np.array([(self.MIN_DEPTH)*np.ones(shape=(len(cc),)), cc[:,0], cc[:,1]]).T

    xs = positions[:,0]
    ys = positions[:,1]
    zs = positions[:,2]

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    ax.scatter(xs, ys, zs)
    ax.set_title("Ideal UAV formation inspection.")

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    
    return positions

  def getIoUsppd(self, positions):
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    xs = positions[:,0]
    ys = positions[:,1]
    zs = positions[:,2]

    ax.scatter(xs, ys, zs, marker='D', label = "Ideal goal positions.")
    for i in range(0, self.nmbr_drones):
      ax.text(xs[i], ys[i], zs[i], str(i), ha='left', c='b')

    while(1):
      print("Minimizing IoU ...")
      combs = combinations(range(0, self.nmbr_drones), 2)
      UAV_IoU = np.zeros(shape=(self.nmbr_drones, 1))

      for i in combs:
        box1 = [self.boxShape, 0, positions[i[0]]/np.array([1, 1, 2])]
        box2 = [self.boxShape, 0, positions[i[1]]/np.array([1, 1, 2])]
        IoU = _3DIoU(box1, box2)

        UAV_IoU[i[0]] += IoU.IOU_3d
        UAV_IoU[i[1]] += IoU.IOU_3d

      maxIoU = max(UAV_IoU)
      maxUAV = np.where(UAV_IoU == maxIoU)[0][0]
      print("Maximum IoU: " + str(maxIoU))

      if maxIoU != 0:
        if max(positions[:,0]) <= self.MAX_DEPTH:
          positions[maxUAV][0] += self.boxShape[0]
        else:
          print("Formation for " + str(self.nmbr_drones) + " UAVs failed at UAV " + str(maxUAV) + " depth limitations exceeded, please lower the number of UAVs.")
          break
      elif maxIoU ==0:
        print("Interception free formation found.")
        break    

    xs = positions[:,0]
    ys = positions[:,1]
    zs = positions[:,2]

    ax.scatter(xs, ys, zs, marker='s', label = "Adjusted goal positions.")
    for i in range(0, self.nmbr_drones):
      ax.text(xs[i], ys[i], zs[i], str(i), ha='left', c='orange')
    ax.legend()

  def descend(positions):
    pass

  def save(self, positions, in_path, out_path):
    file_name = os.path.basename(in_path)
    name = file_name.split('.', 1)[0]
    
    np.savetxt(out_path + name +
                '_mdf_pos.csv', positions, delimiter=",")
    


#!/usr/bin/env python3

import datetime
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from crazyKhoreia.crazyKhoreia import crazyKhoreia


class lightPainting():
    def __init__(self, MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, in_path, out_path, detail=0.05, speed=1.0, sleepTime=1.5, video=False, led=False):

        self.MAX_WIDTH, self.MAX_HEIGHT, self.MIN_WIDTH, self.MIN_HEIGHT, self.in_path, self.out_path = MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, in_path, out_path
        self.detail, self.speed, self.sleepTime, self.video, self.led = detail, speed, sleepTime, video, led

        self.ck = crazyKhoreia(self.MAX_WIDTH, self.MAX_HEIGHT,
                               self.MIN_WIDTH, self.MIN_HEIGHT, self.in_path, self.led)

        self.wayPoints = self.ck.get_waypoints()
        self.wayPoints = self.clean_waypoints(self.wayPoints, self.detail)

        self.distance, self.Time = self.calculate_stats()
        self.msg = self.save(self.wayPoints, self.distance,
                             self.Time, self.in_path, self.out_path, self.video)

        self.ck.plot_contour_inspection(self.wayPoints)

        plt.show()

    def clean_waypoints(self, wayPoints, detail):
        # Clean wayPoints by a detail parameter.
        initialPoints = len(wayPoints)

        for j in range(0, 10):
            for i in range(1, initialPoints):
                if ((i + 1 == len(wayPoints)) | (i == len(wayPoints))):
                    break
                elif (np.linalg.norm(wayPoints[:, 0:3][i - 1] - wayPoints[:, 0:3][i]) < detail) & (wayPoints[:, 3][i] == 1):
                    wayPoints = np.delete(wayPoints, i, 0)

        return wayPoints

    def update(self, num, x, y, line):
        line.set_data(x[:num], y[:num])
        return line,

    def calculate_stats(self):
        distance = 0
        takeOffHeight = self.wayPoints[0][2]
        initialPos = np.array(
            self.wayPoints[:, 0:3][0] - np.array([0, 0, takeOffHeight]))

        for i in range(0, len(self.wayPoints)):
            if i == 0:
                distance += abs(np.linalg.norm(initialPos -
                                self.wayPoints[:, 0:3][0]))
            else:
                distance += abs(np.linalg.norm(
                    self.wayPoints[:, 0:3][i - 1] - self.wayPoints[:, 0:3][i]))

        Time = distance/self.speed*self.sleepTime

        return distance, Time

    def save(self, wayPoints, distance, Time, in_path, out_path, video):
        file_name = os.path.basename(in_path)
        name = file_name.split('.', 1)[0]

        X = wayPoints[:, 0]
        Y = wayPoints[:, 1]
        Z = wayPoints[:, 2]

        # Plot points.
        fig2, ax1 = plt.subplots()
        ax1.plot(wayPoints[:, 1], wayPoints[:, 2], 'o',
                 c='blueviolet', label="waypoints.")
        ax1.plot(Y, Z, color='#570861')
        ax1.set_title("UAV path.")

        fig, ax = plt.subplots()
        line, = ax.plot(Y, Z, color='#570861')
        ax.set_title("UAV path animation.")

        if video == True:
            ani = animation.FuncAnimation(fig, self.update, len(Y), fargs=[Y, Z, line],
                                          interval=50, blit=True)
            ani.save(out_path + name + '_lp_video.mp4')
            plt.show()

        np.savetxt(out_path + name +
                   '_lp_wpts.csv', wayPoints, delimiter=",")

        minCoords = np.array([min(X), min(Y), min(Z)])
        maxCoords = np.array([max(X), max(Y), max(Z)])
        takeOffHeight = wayPoints[0][2]

        initialPos = wayPoints[:, 0:3][0] - np.array([0, 0, takeOffHeight])

        msg = "Choreography ready!, please read the following information:" + \
              "\nInitial position: " + str(initialPos) + \
              "\nTake off heigth: " + str(takeOffHeight) + \
              "\nMinimum coordinates: " + str(minCoords) + \
              "\nMaximum coordinates: " + str(maxCoords) + \
              "\nNumber of waypoints: " + str(len(wayPoints)) + \
              "\nTotal distance: " + str(distance) + " meters." + \
              "\nTotal time: " + str(datetime.timedelta(seconds=Time))

        print(msg)

        return msg



#!/usr/bin/env python3

import cv2 as cv
import numpy as np
from matplotlib import pyplot as plt
from cycler import cycler


class crazyKhoreia():
    def __init__(self, MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, in_path, led=False):
        self.MAX_WIDTH, self.MAX_HEIGHT, self.MIN_WIDTH, self.MIN_HEIGHT, self.in_path, self.led = MAX_WIDTH, MAX_HEIGHT, MIN_WIDTH, MIN_HEIGHT, in_path, led

        # Read image.
        self.img = cv.imread(self.in_path, cv.IMREAD_UNCHANGED)

        # Proccess image and get contours from it.
        contours = self.process_image()

        # Proccess the contours and get its parameters relative to the input image.
        self.cnt_scaled, self.ImgShape_X, self.ImgShape_Y = self.process_contours(
            contours)

    def process_image(self):
        # If the image type is png, it'll have a fourth channel known as alpha, which is transparency,
        # in that case, the background will be filled in white.
        if self.img.shape[2] == 4:
            bg = np.array([255, 255, 255])
            alpha = (self.img[:, :, 3] /
                     255).reshape(self.img.shape[:2] + (1,))
            self.img = ((bg * (1 - alpha)) +
                        (self.img[:, :, :3] * alpha)).astype(np.uint8)

        # Flip image.
        img_flipped = cv.flip(self.img, 0)

        # Invert the input image before detecting contours (see: https://stackoverflow.com/questions/29329866/how-to-avoid-detecting-image-frame-when-using-findcontours).
        img_flipped = 255 - img_flipped

        # Convert the image from BGR to grayscale.
        im_gray = cv.cvtColor(img_flipped, cv.COLOR_BGR2GRAY)

        # Binarize the grayscale image.
        th, img_bw = cv.threshold(im_gray, 128, 192, cv.THRESH_OTSU)

        # Find countours.
        contours, hierarchy = cv.findContours(
            image=img_bw, mode=cv.RETR_TREE, method=cv.CHAIN_APPROX_NONE)

        # Create subplots for original image and image with contours visualization.
        fig, (ax0, ax1, ax2) = plt.subplots(nrows=1, ncols=3)

        # img[..., ::-1] reverts the image's channel order from BGR to RGB so it can be correctly displayed by plt.imshow()
        ax0.imshow(self.img[..., ::-1])
        ax0.set_axis_off()
        ax0.set_title("Original image.")
        ax1.imshow(cv.flip(img_bw, 0), cmap='gray')
        ax1.set_axis_off()
        ax1.set_title("Threshold image.")
        ax2.set_axis_off()
        ax2.set_title("Image with contours.")

        # Draw contours on original image.
        img_with_contour = cv.drawContours(
            image=255 - img_flipped, contours=contours, contourIdx=-1, color=(0, 255, 0), thickness=3, lineType=cv.LINE_AA)

        ax2.imshow(cv.flip(img_with_contour[..., ::-1], 0))

        return contours

    def process_contours(self, contours):
        # Get image's dimensions.
        ImgShape_Y, ImgShape_X = self.img.shape[:2]

        # Get image's scale factor from its dimensions and user's parameters.
        scale_fact_x = ImgShape_X/(self.MAX_WIDTH - self.MIN_WIDTH)
        scale_fact_y = ImgShape_Y/(self.MAX_HEIGHT - self.MIN_HEIGHT)

        # Scale contours if needed to satisfy maximum dimensions requirements.
        cnt_scaled = []

        for cnt in contours:
            cnt = cnt - [ImgShape_X/2, ImgShape_Y/2]
            if(scale_fact_x > 1.0 or scale_fact_y > 1.0):
                if(scale_fact_x > scale_fact_y):
                    cnt_scaled.append(cnt*(1.0/scale_fact_x))
                else:
                    cnt_scaled.append(cnt*(1.0/scale_fact_y))

        # Adjust image parameters relative to its scaled contours.
        if(scale_fact_x > scale_fact_y):
            ImgShape_X /= scale_fact_x
            ImgShape_Y /= scale_fact_x
        if(scale_fact_y > scale_fact_x):
            ImgShape_X /= scale_fact_y
            ImgShape_Y /= scale_fact_y

       # Translate the scaled contours to satisfy minimum dimensions requirements.
        minX = []
        minY = []

        for cnt in cnt_scaled:
            x = np.array(cnt)
            x = np.reshape(x, (len(cnt), 2))
            minX.append(min(x[:, 0]))
            minY.append(min(x[:, 1]))

        minX = abs(min(minX))
        minY = abs(min(minY))

        cnt_scaled_x = []
        for cnt in cnt_scaled:
            cnt_x = cnt + [minX, minY] + [self.MIN_WIDTH, self.MIN_HEIGHT]
            cnt_scaled_x.append(cnt_x)

        cnt_scaled = cnt_scaled_x

        return cnt_scaled, ImgShape_X, ImgShape_Y

    def plot_contour_inspection(self, wayPoints):
        strPoints = np.empty((0, 2))
        endPoints = np.empty((0, 2))

        # See https://matplotlib.org/3.5.1/gallery/color/named_colors.html#sphx-glr-gallery-color-named-colors-py for more colors.
        cc = (cycler(color=['purple', 'orange', 'lime', 'royalblue', 'steelblue', 'cyan', 'gold']) +
              cycler(marker=['^', 'o', 's', 'p', '*', 'x', 'D']))

        fig, ax = plt.subplots()
        ax.set_prop_cycle(cc)

        # Iterate contours.
        for i in range(0, len(self.cnt_scaled)):
            x = np.array(self.cnt_scaled[i])
            x = np.reshape(x, (len(self.cnt_scaled[i]), 2))

            strPoints = np.vstack([strPoints, [x[0, 0], x[0, 1]]])
            endPoints = np.vstack([endPoints, [x[-1, 0], x[-1, 1]]])

            ax.plot(x[:, 0], x[:, 1], label="Contour " + str(i), ls='--')

        ax.plot(strPoints[:, 0], strPoints[:, 1], 'o', c='b',
                label="Start points.", alpha=0.30, ms=12)
        ax.plot(endPoints[:, 0], endPoints[:, 1], 's', c='r',
                label="End points.", alpha=0.30, ms=12)

        for i in range(0, len(strPoints)):
            ax.arrow(strPoints[:, 0][i], strPoints[:, 1][i], endPoints[:, 0][i] -
                     strPoints[:, 0][i], endPoints[:, 1][i] - strPoints[:, 1][i], ls='--', color='g')
            ax.text(strPoints[:, 0][i], strPoints[:, 1]
                    [i], str(i), ha='left', c='b')
            ax.text(endPoints[:, 0][i], endPoints[:, 1]
                    [i], str(i), ha='right', c='r')

        ax.plot(wayPoints[:, 1], wayPoints[:, 2], c='silver',
                label='wayPoints', alpha=0.8, ls=':', marker=',')
        ax.legend()
        ax.set_title("Contour inspection.")

    def get_waypoints(self):
        # Convert vectors (contours) to cartesian points in XYZ [meters] format.
        if self.led == True:
            wayPoints = np.empty((0, 4))
        else:
            wayPoints = np.empty((0, 3))

        # Iterate contours.
        for i in range(0, len(self.cnt_scaled)):
            x = np.array(self.cnt_scaled[i])
            x = np.reshape(x, (len(self.cnt_scaled[i]), 2))

            if self.led == True:
                led = np.ones(shape=(len(x),))
                led[0] = 0
                stck = np.array(
                    [1.5*np.ones(shape=(len(x),)), x[:, 0], x[:, 1], led]).T
            else:
                stck = np.array(
                    [1.5*np.ones(shape=(len(x),)), x[:, 0], x[:, 1]]).T

            wayPoints = np.vstack([wayPoints, stck])

        return wayPoints


#!/usr/bin/env python3

# 3D IoU caculate code for 3D object detection
# Kent 2018/12

import numpy as np
from scipy.spatial import ConvexHull
from numpy import *


class _3DIoU():
    """ Calculate the Interception over the Union between a pair of 3D boxes.
    """

    def __init__(self, box1, box2):
        """ Each box consists of a list containing:
        box --> [box_size: tuple of (length (X axis), wide (Y axis), height (Z axis)),
                 heading_angle: rad scalar, clockwise from pos x axis,
                 center: tuple of (x, y, z)]
        """
        self.box1 = box1
        self.box2 = box2

        self.corners1 = self.get_3d_box(
            self.box1[0], self.box1[1], self.box1[2])
        self.corners2 = self.get_3d_box(
            self.box2[0], self.box2[1], self.box2[2])

        (self.IOU_3d, self.IOU_2d) = self.box3d_iou(
            self.corners2, self.corners1)

    def polygon_clip(self, subjectPolygon, clipPolygon):
        """ Clip a polygon with another polygon.
        Ref: https://rosettacode.org/wiki/Sutherland-Hodgman_polygon_clipping#Python
        Args:
          subjectPolygon: a list of (x,y) 2d points, any polygon.
          clipPolygon: a list of (x,y) 2d points, has to be *convex*
        Note:
          **points have to be counter-clockwise ordered**
        Return:
          a list of (x,y) vertex point for the intersection polygon.
        """
        def inside(p):
            return(cp2[0]-cp1[0])*(p[1]-cp1[1]) > (cp2[1]-cp1[1])*(p[0]-cp1[0])

        def computeIntersection():
            dc = [cp1[0] - cp2[0], cp1[1] - cp2[1]]
            dp = [s[0] - e[0], s[1] - e[1]]
            n1 = cp1[0] * cp2[1] - cp1[1] * cp2[0]
            n2 = s[0] * e[1] - s[1] * e[0]
            n3 = 1.0 / (dc[0] * dp[1] - dc[1] * dp[0])
            return [(n1*dp[0] - n2*dc[0]) * n3, (n1*dp[1] - n2*dc[1]) * n3]

        outputList = subjectPolygon
        cp1 = clipPolygon[-1]

        for clipVertex in clipPolygon:
            cp2 = clipVertex
            inputList = outputList
            outputList = []
            s = inputList[-1]

            for subjectVertex in inputList:
                e = subjectVertex
                if inside(e):
                    if not inside(s):
                        outputList.append(computeIntersection())
                    outputList.append(e)
                elif inside(s):
                    outputList.append(computeIntersection())
                s = e
            cp1 = cp2
            if len(outputList) == 0:
                return None
        return(outputList)

    def poly_area(self, x, y):
        """ Ref: http://stackoverflow.com/questions/24467972/calculate-area-of-polygon-given-x-y-coordinates """
        return 0.5*np.abs(np.dot(x, np.roll(y, 1))-np.dot(y, np.roll(x, 1)))

    def convex_hull_intersection(self, p1, p2):
        """ Compute area of two convex hull's intersection area.
            p1,p2 are a list of (x,y) tuples of hull vertices.
            return a list of (x,y) for the intersection and its volume
        """
        inter_p = self.polygon_clip(p1, p2)
        if inter_p is not None:
            hull_inter = ConvexHull(inter_p, qhull_options='QJ')
            return inter_p, hull_inter.volume
        else:
            return None, 0.0

    def box3d_vol(self, corners):
        ''' corners: (8,3) no assumption on axis direction '''
        a = np.sqrt(np.sum((corners[0, :] - corners[1, :])**2))
        b = np.sqrt(np.sum((corners[1, :] - corners[2, :])**2))
        c = np.sqrt(np.sum((corners[0, :] - corners[4, :])**2))
        return a*b*c

    def is_clockwise(self, p):
        x = p[:, 0]
        y = p[:, 1]
        return np.dot(x, np.roll(y, 1))-np.dot(y, np.roll(x, 1)) > 0

    def box3d_iou(self, corners1, corners2):
        ''' Compute 3D bounding box IoU.
        Input:
            corners1: numpy array (8,3), assume up direction is negative Y
            corners2: numpy array (8,3), assume up direction is negative Y
        Output:
            iou: 3D bounding box IoU
            iou_2d: bird's eye view 2D bounding box IoU
        todo (kent): add more description on corner points' orders.
        '''
        # corner points are in counter clockwise order
        rect1 = [(corners1[i, 0], corners1[i, 2]) for i in range(3, -1, -1)]
        rect2 = [(corners2[i, 0], corners2[i, 2]) for i in range(3, -1, -1)]

        area1 = self.poly_area(np.array(rect1)[:, 0], np.array(rect1)[:, 1])
        area2 = self.poly_area(np.array(rect2)[:, 0], np.array(rect2)[:, 1])

        inter, inter_area = self.convex_hull_intersection(rect1, rect2)
        iou_2d = inter_area/(area1+area2-inter_area)
        ymax = min(corners1[0, 1], corners2[0, 1])
        ymin = max(corners1[4, 1], corners2[4, 1])

        inter_vol = inter_area * max(0.0, ymax-ymin)

        vol1 = self.box3d_vol(corners1)
        vol2 = self.box3d_vol(corners2)
        iou = inter_vol / (vol1 + vol2 - inter_vol)
        return iou, iou_2d

    # ----------------------------------
    # Helper functions for evaluation
    # ----------------------------------

    def get_3d_box(self, box_size, heading_angle, center):
        ''' Calculate 3D bounding box corners from its parameterization.
        Input:
            box_size: tuple of (length,wide,height)
            heading_angle: rad scalar, clockwise from pos x axis
            center: tuple of (x,y,z)
        Output:
            corners_3d: numpy array of shape (8,3) for 3D box cornders
        '''
        def roty(t):
            c = np.cos(t)
            s = np.sin(t)
            return np.array([[c,  0,  s],
                            [0,  1,  0],
                            [-s, 0,  c]])

        R = roty(heading_angle)
        l, h, w = box_size
        x_corners = [l/2, l/2, -l/2, -l/2, l/2, l/2, -l/2, -l/2]
        y_corners = [h/2, h/2, h/2, h/2, -h/2, -h/2, -h/2, -h/2]
        z_corners = [w/2, -w/2, -w/2, w/2, w/2, -w/2, -w/2, w/2]
        corners_3d = np.dot(R, np.vstack([x_corners, y_corners, z_corners]))
        corners_3d[0, :] = corners_3d[0, :] + center[0]
        corners_3d[1, :] = corners_3d[1, :] + center[1]
        corners_3d[2, :] = corners_3d[2, :] + center[2]
        corners_3d = np.transpose(corners_3d)
        return corners_3d
